/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QFile>
#include <QFileInfo>
#include <QTextStream>
#include <cmath>
#include <assert.h>

#include "ballbarpattern.h"

TBallBarPattern::TBallBarPattern(void)
{
	d_pattern_average = 0.0;
	d_pattern_stdev = 0.0;
	d_pattern_min_length = 0.0;
	d_pattern_max_length = 0.0;
}

TBallBarPattern::~TBallBarPattern(void)
{

}

TVector3 TBallBarPattern::Volume_Corner_Min(void) const
{
	TVector3							pnt(-100,-100,-100);
	bool								init(true);
	std::vector<TBallBarPattern::TBallBarMeasurement>::const_iterator iter;
	
	for(iter = d_ballbar_measurements.begin();iter != d_ballbar_measurements.end();++iter)
	{
		if(init)
		{
			pnt = (*iter).sphere_position1;
			init = false;
		}
		
		if((*iter).sphere_position1.x < pnt.x)
		{
			pnt.x = (*iter).sphere_position1.x;
		}
		
		if((*iter).sphere_position2.x < pnt.x)
		{
			pnt.x = (*iter).sphere_position2.x;
		}
		
		if((*iter).sphere_position1.y < pnt.y)
		{
			pnt.y = (*iter).sphere_position1.y;
		}
		
		if((*iter).sphere_position2.y < pnt.y)
		{
			pnt.y = (*iter).sphere_position2.y;
		}
		
		if((*iter).sphere_position1.z < pnt.z)
		{
			pnt.z = (*iter).sphere_position1.z;
		}
		
		if((*iter).sphere_position2.z < pnt.z)
		{
			pnt.z = (*iter).sphere_position2.z;
		}
	}
	
	return pnt;
}

TVector3 TBallBarPattern::Volume_Corner_Max(void) const
{
	TVector3							pnt(100,100,100);
	bool								init(true);
	std::vector<TBallBarPattern::TBallBarMeasurement>::const_iterator iter;
	
	for(iter = d_ballbar_measurements.begin();iter != d_ballbar_measurements.end();++iter)
	{
		if(init)
		{
			pnt = (*iter).sphere_position1;
			init = false;
		}
		
		if((*iter).sphere_position1.x > pnt.x)
		{
			pnt.x = (*iter).sphere_position1.x;
		}
		
		if((*iter).sphere_position2.x > pnt.x)
		{
			pnt.x = (*iter).sphere_position2.x;
		}
		
		if((*iter).sphere_position1.y > pnt.y)
		{
			pnt.y = (*iter).sphere_position1.y;
		}
		
		if((*iter).sphere_position2.y > pnt.y)
		{
			pnt.y = (*iter).sphere_position2.y;
		}
		
		if((*iter).sphere_position1.z > pnt.z)
		{
			pnt.z = (*iter).sphere_position1.z;
		}
		
		if((*iter).sphere_position2.z > pnt.z)
		{
			pnt.z = (*iter).sphere_position2.z;
		}
	}
	
	return pnt;
}

TBallBarPattern::TBallBarMeasurement TBallBarPattern::BallBar_Measurement(
	const int 							index) const
{
	assert(static_cast<unsigned int>(index) < d_ballbar_measurements.size());

	return d_ballbar_measurements[index];
}

bool TBallBarPattern::Load_Data(
	const QString						&file_name)
{
	QFile								file;
	QByteArray							file_data;
	TBallBarPattern::TFileEncoding		file_encoding;
	TBallBarPattern::TFileContents		file_contents;
	
	this->Clear();
	
	file.setFileName(file_name);
	
	if(file.open(QIODevice::ReadOnly))
	{
		file_data = file.readAll();
		file.close();
	}
	else
	{
		d_last_error = QString("Cannot open file '%1'").arg(file_name);
		return false;
	}
	
	file_encoding = this->DetermineFileEncoding(file_data);
	
	if(file_encoding != TBallBarPattern::ASCII_FILE_TYPE)
	{
		d_last_error = QString("Input file '%1' appears to be binary and is not supported.").arg(file_name);
		return false;
	}
	
	file_contents = this->DetermineAsciiFileContents(file_data);
	
	if(file_contents == TBallBarPattern::UNKNOWN_CONTENTS)
	{
		d_last_error = QString("Input file '%1' is not recognized as a ball bar measurement file.").arg(file_name);
		return false;
	}
	
	switch(file_contents)
	{
		case TBallBarPattern::BALLBAR_DATAFILE_LEGACY_V1:
			if(!this->Load_Data_Legacy_V1(file_data))
			{
				return false;
			}
			break;
			
		case TBallBarPattern::BALLBAR_DATAFILE_LEGACY_V2:
			if(!this->Load_Data_Legacy_V2(file_data))
			{
				return false;
			}
			break;
			
		case TBallBarPattern::BALLBAR_DATAFILE_V1:
			if(!this->Load_Data_V1(file_data))
			{
				return false;
			}
			break;
			
		case TBallBarPattern::UNKNOWN_CONTENTS:
			assert(false);				// should never get here
			break;			
	}

	
	this->Update_Statistics();
	
	return true;
}

TBallBarPattern::TFileEncoding TBallBarPattern::DetermineFileEncoding(
	const QByteArray					&file_data) const
{
	TBallBarPattern::TFileEncoding			file_encoding(TBallBarPattern::ASCII_FILE_TYPE);
	int									cntr;
	
	for(cntr = 0;cntr < file_data.size();++cntr)
	{
		if(file_data[cntr] != 0x0d &&	// cr
		   file_data[cntr] != 0x0a &&	// lf
		   file_data[cntr] != '\t' &&	// horizontal tab
		   file_data[cntr] != '\v' &&	// vertical tab
		   file_data[cntr] < 0x20)
		{
			file_encoding = TBallBarPattern::BINARY_FILE_TYPE;
			break;
		}
	}

	return file_encoding;
}

TBallBarPattern::TFileContents TBallBarPattern::DetermineAsciiFileContents(
	const QByteArray					&file_data) const
{
	TBallBarPattern::TFileContents			file_contents(TBallBarPattern::UNKNOWN_CONTENTS);
	QString								file_text(file_data);
	int									index;
	
	if(file_text.length() > 10)				// cannot be smaller than this
	{
		file_text = file_text.trimmed();	// remove all leading spaces
		
		index = file_text.indexOf('\n');
		
		if(!(index < 0))
		{
			file_text.truncate(index);
		}
		
		if(file_text.startsWith(QStringLiteral("BallBar_V1"),Qt::CaseInsensitive))
		{
			file_contents = TBallBarPattern::BALLBAR_DATAFILE_LEGACY_V1;
		}
		else if(file_text.startsWith(QStringLiteral("BallBar_V2"),Qt::CaseInsensitive))
		{
			file_contents = TBallBarPattern::BALLBAR_DATAFILE_LEGACY_V2;
		}
		else if(file_text.startsWith(QStringLiteral("B89.4.1_Data:Version=1:Type=BallBar"),Qt::CaseInsensitive))
		{
			file_contents = TBallBarPattern::BALLBAR_DATAFILE_V1;
		}
	}
	
	return file_contents;
}

void TBallBarPattern::Clear(void)
{
	d_ballbar_measurements.clear();
	this->Update_Statistics();
}

void TBallBarPattern::Update_Statistics(void)
{
	std::vector<TBallBarPattern::TBallBarMeasurement>::const_iterator iter;
	bool								init;
	double								length;
	int									count;
	double								sq_dev(0);
	
	init = true;

	d_pattern_average = 0.0;
	d_pattern_stdev = 0.0;
	d_pattern_min_length = 0.0;
	d_pattern_max_length = 0.0;
	
	count = static_cast<int>(d_ballbar_measurements.size());
	
	if(count)
	{
		for(iter = d_ballbar_measurements.begin();iter != d_ballbar_measurements.end();++iter)
		{
			length = (*iter).length;
			
			if(init)
			{
				d_pattern_min_length = d_pattern_max_length = length;
				init = false;
			}
			else
			{
				if(length > d_pattern_max_length)
				{
					d_pattern_max_length = length;
				}
				else if(length < d_pattern_min_length)
				{
					d_pattern_min_length = length;
				}
			}
			
			d_pattern_average += length;
		}

		d_pattern_average /= static_cast<double>(count);

		if(count > 1)
		{
			for(iter = d_ballbar_measurements.begin();iter != d_ballbar_measurements.end();++iter)
			{
				sq_dev += pow(((*iter).length - d_pattern_average),2);
			}

			sq_dev /= static_cast<double>(count - 1);
			d_pattern_stdev = sqrt(sq_dev);
		}
	}
}

bool TBallBarPattern::Load_Data_Legacy_V1(
	const QByteArray					&file_data)
{
	QTextStream							text_stream(file_data,QIODevice::ReadOnly);
	QString								text_line;
	QStringList							text_values;
	int									count;
	TBallBarPattern::TBallBarMeasurement meas;
	bool								conversion_state;

	while(!text_stream.atEnd())
	{
		text_line = text_stream.readLine().simplified();
		
		if(text_line.length())
		{
			if(!(text_line.startsWith(QStringLiteral("#"),Qt::CaseInsensitive) ||
				 text_line.startsWith(QStringLiteral("BallBar_V1"),Qt::CaseInsensitive)))
			{
				text_values = text_line.split(',');
				
				count = text_values.size();
				
				if(count == 11)
				{
					meas.name = text_values[0];
					meas.length = text_values[1].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
			
					meas.sphere_position1.x = text_values[2].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.sphere_position1.y = text_values[3].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.sphere_position1.z = text_values[4].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					
					meas.sphere_position2.x = text_values[5].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.sphere_position2.y = text_values[6].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.sphere_position2.z = text_values[7].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					
					meas.probe_offset.x = text_values[8].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.probe_offset.y = text_values[9].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.probe_offset.z = text_values[10].toDouble(&conversion_state);
					if(!conversion_state) continue;
					

					// sphere positions for V1 files in map coordinates
					// must be converted to true measurement position
					
					meas.sphere_position1 += meas.probe_offset;
					meas.sphere_position2 += meas.probe_offset;
					
					// unused fields
					meas.sphere_form1 = 0.0;
					meas.sphere_form2 = 0.0;
					
					d_ballbar_measurements.push_back(meas);
				}
			}
		}
	}
	
	return true;
}

bool TBallBarPattern::Load_Data_Legacy_V2(
	const QByteArray					&file_data)
{
	QTextStream							text_stream(file_data,QIODevice::ReadOnly);
	QString								text_line;
	QStringList							text_values;
	int									count;
	TBallBarPattern::TBallBarMeasurement meas;
	bool								conversion_state;
	
	while(!text_stream.atEnd())
	{
		text_line = text_stream.readLine().simplified();
		
		if(text_line.length())
		{
			if(!(text_line.startsWith(QStringLiteral("#"),Qt::CaseInsensitive) ||
				 text_line.startsWith(QStringLiteral("BallBar_V2"),Qt::CaseInsensitive)))
			{
				text_values = text_line.split(',');
				
				count = text_values.size();
				
				if(count == 13)
				{
					meas.name = text_values[0];
					meas.length = text_values[1].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
		
					meas.sphere_position1.x = text_values[2].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.sphere_position1.y = text_values[3].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.sphere_position1.z = text_values[4].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					
					meas.sphere_position2.x = text_values[5].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.sphere_position2.y = text_values[6].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.sphere_position2.z = text_values[7].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					
					meas.sphere_form1 = text_values[8].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.sphere_form2 = text_values[9].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					
					meas.probe_offset.x = text_values[10].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.probe_offset.y = text_values[11].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.probe_offset.z = text_values[12].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					
					// sphere positions for V2 files in map coordinates
					// must be converted to true measurement position
					
					meas.sphere_position1 += meas.probe_offset;
					meas.sphere_position2 += meas.probe_offset;
					
					d_ballbar_measurements.push_back(meas);
				}
			}
		}
	}
	
	return true;
}

bool TBallBarPattern::Load_Data_V1(
	const QByteArray					&file_data)
{
	QTextStream							text_stream(file_data,QIODevice::ReadOnly);
	QString								text_line;
	QStringList							text_values;
	int									count;
	bool								conversion_state;
	TBallBarPattern::TBallBarMeasurement meas;
	
	while(!text_stream.atEnd())
	{
		text_line = text_stream.readLine().simplified();
		
		if(text_line.length())
		{
			if(!(text_line.startsWith(QStringLiteral("#"),Qt::CaseInsensitive) ||
				 text_line.startsWith(QStringLiteral("B89.4.1_Data:Version=1:Type=BallBar"),Qt::CaseInsensitive) ||
			   (text_line.contains(QStringLiteral("Name"),Qt::CaseInsensitive) && text_line.contains(QStringLiteral("Length"),Qt::CaseInsensitive))))
			{
				text_values = text_line.split(',');
				
				count = text_values.size();
				
				if(count == 13)
				{
					meas.name = text_values[0];
					meas.length = text_values[1].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.sphere_position1.x = text_values[2].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.sphere_position1.y = text_values[3].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.sphere_position1.z = text_values[4].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					
					meas.sphere_position2.x = text_values[5].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.sphere_position2.y = text_values[6].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.sphere_position2.z = text_values[7].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					
					meas.sphere_form1 = text_values[8].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.sphere_form2 = text_values[9].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					
					meas.probe_offset.x = text_values[10].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.probe_offset.y = text_values[11].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					meas.probe_offset.z = text_values[12].toDouble(&conversion_state);
					if(!conversion_state) continue;
					
					
					// sphere positions for NF ballbar files include probe offsets
					
					d_ballbar_measurements.push_back(meas);
				}
			}
		}
	}
	
	return true;
}




