/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////



#include <QAction>
#include <QApplication>
#include <QGridLayout>
#include <QWidget>
#include <QLabel>
#include <QFrame>
#include <QTabWidget>
#include <QPlainTextEdit>
#include <QSplitter>
#include <QDragEnterEvent>
#include <QDropEvent>
#include <QMimeData>
#include <QUrl>
#include <QDockWidget>
#include <QMenu>
#include <QMenuBar>
#include <QToolBar>
#include <QSpacerItem>
#include <QToolButton>
#include <QScreen>
#include <QSettings>
#include <QTimer>
#include <QPixmap>
#include <QSizePolicy>
#include <QFont>
#include <QDir>
#include <QFile>
#include <QFileDialog>
#include <QList>
#include <QRect>
#include <QSize>
#include <QPoint>

#include "../../core/messagebox.h"
#include "../../core/graphwidget.h"

#include "tolerancewidget.h"
#include "measurementwidget.h"
#include "opengldisplaywidget.h"

#include "ballbarviewer.h"


TBallBarViewer::TBallBarViewer(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QMainWindow(const_cast<QWidget*>(parent),flags)
{
	QTabWidget							*information_tab;
	QDockWidget							*measurement_dock;
	QDockWidget							*tolerance_dock;
	QGridLayout							*grid_layout;
	QGridLayout							*info_layout;
	QLabel								*logo_label;
	QLabel								*copyright_label;
	QMenu								*file_menu;
	QMenu								*view_menu;
	QMenu								*help_menu;
	QMenuBar							*menubar;
	QWidget								*central_widget;
	QWidget								*info_widget;
	QToolBar							*tool_bar;
	QSizePolicy							sizepolicy_preferred_minimum(QSizePolicy::Preferred, QSizePolicy::Minimum);
	QRect								desktop_rect;
	QRect								window_rect;
	QPoint								position;
	QSize								size;
	QDir								data_path;
	QList<int>							splitter_sizes;
	TVector3							min_machine;
	TVector3							max_machine;
	TVector3							move_position;
	TVector3							center_machine;
	QFont								log_font(QStringLiteral("courier new"));
	QFont								copyright_font;
	QFont								text_font;
	QString								text;
	
	this->setAcceptDrops(true);

	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);
	
	text_font.setFamily(QStringLiteral("Courier New"));
	text_font.setFixedPitch(true);
	
	sizepolicy_preferred_minimum.setHorizontalStretch(0);
	sizepolicy_preferred_minimum.setVerticalStretch(0);

	d_application_path = qApp->applicationDirPath();
	log_font.setFixedPitch(true);

#ifdef Q_OS_MAC
	int index = d_application_path.lastIndexOf(QStringLiteral(".app/"));
	d_application_path.truncate(index);
	
	index = d_application_path.lastIndexOf('/');
	d_application_path.truncate(index);
#endif
	
#ifdef Q_OS_WIN
	d_data_path = d_application_path;
#else
	d_data_path = QDir::homePath();
#endif

	data_path.setPath(d_data_path);
	
	if(!data_path.exists(QStringLiteral(".ballbarviewer")))
	{
		data_path.mkdir(QStringLiteral(".ballbarviewer"));
	}
	
	data_path.cd(QStringLiteral(".ballbarviewer"));
	d_data_path = data_path.absolutePath();
	
	d_settings = new QSettings(d_data_path + QStringLiteral("/ballbarviewer_settings.ini"),QSettings::IniFormat,this);
	
	position = d_settings->value("Mainwindow_Position", QPoint(30,30)).toPoint();
	size = d_settings->value("Mainwindow_Size", QSize(1001, 801)).toSize();
	
	desktop_rect = QGuiApplication::primaryScreen()->availableGeometry();
	
	window_rect.moveTopLeft(position);
	window_rect.setSize(size);
	
	window_rect = window_rect & desktop_rect;
	
	if(window_rect.isValid() && window_rect.width() > 500 && window_rect.height() > 500)
	{
		this->resize(window_rect.size());
		this->move(window_rect.topLeft());
	}
	else
	{
		this->resize(1001, 801);
		this->move(QPoint(30,30));
	}
	
	splitter_sizes.push_back(d_settings->value("Graphic_Window_Height", 800).toInt());
	splitter_sizes.push_back(d_settings->value("Analysis_Window_Height", 400).toInt());

	d_file_new_action = new QAction(this);
	d_file_new_action->setShortcut(QKeySequence::New);
	d_file_new_action->setIcon(QIcon(":/menu/file_new.png"));

	d_file_import_ballbar_action = new QAction(this);
	d_file_import_ballbar_action->setIcon(QIcon(":/menu/file_import_ballbar.png"));

	d_file_quit_action = new QAction(this);
	d_file_quit_action->setShortcut(QKeySequence::Quit);

	d_help_about_action = new QAction(this);
	d_help_about_action->setIcon(QIcon(":/menu/help_about.png"));
	
	tool_bar = new QToolBar(this);
	this->addToolBar(Qt::TopToolBarArea, tool_bar);

	central_widget = new QWidget(this);
	this->setCentralWidget(central_widget);
	
	grid_layout = new QGridLayout(central_widget);

	d_splitter_widget = new QSplitter(Qt::Vertical,central_widget);
	grid_layout->addWidget(d_splitter_widget);

	d_opengl_display_widget = new TOpenGLDisplayWidget();
	d_splitter_widget->addWidget(d_opengl_display_widget);
	
	info_widget = new QWidget();
	info_layout = new QGridLayout(info_widget);
	info_layout->setContentsMargins(0,0,0,0);
	
	information_tab = new QTabWidget(info_widget);
	information_tab->setTabPosition(QTabWidget::West);
	
	d_analysis_graph_widget = new TGraphWidget();
	d_analysis_graph_widget->setMinimumSize(0,100);
	d_analysis_graph_widget->SetGraphType(TGraphWidget::BAR_GRAPH);
	d_analysis_graph_widget->SetHLegendMode(TGraphWidget::DEFAULT_LEGEND_PRECISION);
	d_analysis_graph_widget->SetVLegendMode(TGraphWidget::INTEGER_LEGEND_PRECISION);
	d_analysis_graph_widget->Enable_Direct_Render(true);
	information_tab->addTab(d_analysis_graph_widget,QStringLiteral("Histogram"));
	
	d_analysis_text_edit = new QPlainTextEdit();
	d_analysis_text_edit->setMinimumSize(0,0);
	d_analysis_text_edit->setFont(text_font);
	d_analysis_text_edit->setLineWrapMode(QPlainTextEdit::NoWrap);
	d_analysis_text_edit->setReadOnly(true);
	information_tab->addTab(d_analysis_text_edit,QStringLiteral("Analysis"));

	info_layout->addWidget(information_tab,0,0,1,2);

	logo_label = new QLabel(info_widget);
	logo_label->setMinimumSize(100,40);
	logo_label->setMaximumSize(100,40);
	info_layout->addWidget(logo_label,1,1,1,1);
	
	copyright_label = new QLabel(info_widget);
	copyright_label->setFont(copyright_font);
	sizepolicy_preferred_minimum.setHeightForWidth(copyright_label->sizePolicy().hasHeightForWidth());
	copyright_label->setSizePolicy(sizepolicy_preferred_minimum);
	copyright_label->setAlignment(Qt::AlignBottom|Qt::AlignLeft);	// Qt::AlignRight|Qt::AlignTrailing
	info_layout->addWidget(copyright_label,1,0,1,1);

	d_splitter_widget->addWidget(info_widget);

	menubar = new QMenuBar(this);
	this->setMenuBar(menubar);

	file_menu = new QMenu(menubar);
	
	file_menu->addAction(d_file_new_action);
	file_menu->addSeparator();
	file_menu->addAction(d_file_import_ballbar_action);
	file_menu->addSeparator();
	file_menu->addAction(d_file_quit_action);
	
	menubar->addMenu(file_menu);
	
	view_menu = new QMenu(menubar);
	menubar->addMenu(view_menu);

	help_menu = new QMenu(menubar);
	help_menu->addAction(d_help_about_action);
	menubar->resize(1068,22);
	menubar->addMenu(help_menu);

	menubar->addAction(file_menu->menuAction());
	menubar->addAction(view_menu->menuAction());
	menubar->addAction(help_menu->menuAction());
	
	measurement_dock = new QDockWidget(this);
	measurement_dock->setAllowedAreas(Qt::LeftDockWidgetArea|Qt::RightDockWidgetArea);
	this->addDockWidget(Qt::LeftDockWidgetArea,measurement_dock);
	
	d_measurement_widget = new TMeasurementWidget();
	measurement_dock->setWidget(d_measurement_widget);
	
	
	tolerance_dock = new QDockWidget(this);
	tolerance_dock->setAllowedAreas(Qt::LeftDockWidgetArea|Qt::RightDockWidgetArea);
	this->addDockWidget(Qt::LeftDockWidgetArea,tolerance_dock);
	
	d_tolerance_widget = new TToleranceWidget();
	tolerance_dock->setWidget(d_tolerance_widget);
	
	view_menu->addAction(measurement_dock->toggleViewAction());
	view_menu->addAction(tolerance_dock->toggleViewAction());
	
	d_tolerance_widget->Set_Tolerance_A(d_settings->value("Tolerance_A","0.012").toDouble());
	d_tolerance_widget->Set_Tolerance_B(d_settings->value("Tolerance_B","0.006").toDouble());
	d_tolerance_widget->Set_Tolerance_C(d_settings->value("Tolerance_C","1000").toDouble());
	d_tolerance_widget->Set_Tolerance_D(d_settings->value("Tolerance_D","0").toDouble());
	
	d_tolerance_widget->Update();

	d_msg_box = new TMessageBox(this);
	
	// defaults
	d_splitter_widget->setSizes(splitter_sizes);
	
	d_analysis_graph_widget->HorizontalLegend("Ball Bar Length");
	d_analysis_graph_widget->VerticalLegend("Quantity");
	d_analysis_graph_widget->Title("Measurement Distribution");
	d_analysis_graph_widget->EnableMouseCoordinateDisplay(false);
	
	logo_label->setPixmap(QPixmap(QStringLiteral(":/images/logo.png")));
	
	tool_bar->addAction(d_file_new_action);
	tool_bar->addSeparator();
	tool_bar->addAction(d_file_import_ballbar_action);
	tool_bar->addSeparator();
	tool_bar->addAction(d_help_about_action);

	this->setWindowTitle(QStringLiteral("Ballbar Viewer"));
	
	copyright_label->setText(QStringLiteral("Copyright (C) 2024 Select Calibration Incorporated.  www.selectcalibration.ca"));
	d_file_new_action->setText(QStringLiteral("New"));
	d_file_import_ballbar_action->setText(QStringLiteral("Import Ballbar Measurements"));
	d_file_quit_action->setText(QStringLiteral("Quit"));
	d_help_about_action->setText(QStringLiteral("About"));
	file_menu->setTitle(QStringLiteral("File"));
	view_menu->setTitle(QStringLiteral("View"));
	help_menu->setTitle(QStringLiteral("Help"));
	
	measurement_dock->setWindowTitle(QStringLiteral("Measurement"));
	tolerance_dock->setWindowTitle(QStringLiteral("Tolerance"));
	
	connect(d_file_new_action,&QAction::triggered,this,&TBallBarViewer::File_New);
	connect(d_file_import_ballbar_action,&QAction::triggered,this,&TBallBarViewer::File_Import_Ballbar);
	connect(d_help_about_action,&QAction::triggered,this,&TBallBarViewer::Menu_About);
	connect(d_file_quit_action,&QAction::triggered,this,&TBallBarViewer::close);
	
	connect(d_tolerance_widget,&TToleranceWidget::Tolerance_Changed,this,&TBallBarViewer::Tolerance_Changed);
	connect(d_tolerance_widget,&TToleranceWidget::Sub_Tolerance_Changed,this,&TBallBarViewer::Tolerance_Changed);

	connect(d_opengl_display_widget,&TOpenGLDisplayWidget::Ballbar_Item_Clicked,this,&TBallBarViewer::Ballbar_Item_Clicked);
}

TBallBarViewer::~TBallBarViewer(void)
{
}
void TBallBarViewer::File_New(void)
{
	d_opengl_display_widget->Clear_BallBars();
	d_analysis_graph_widget->DeleteDataSets();
	d_analysis_text_edit->clear();
	
	d_opengl_display_widget->Update(false);
	d_analysis_graph_widget->UpdateGraph();
	d_measurement_widget->Enable_Display(false);
	d_tolerance_widget->Enable_Results(false);
}

void TBallBarViewer::File_Import_Ballbar(void)
{
	QString								file_name;
	QString								file_filters;
	
	file_filters = QString("Ballbar Measurement File(*.dat)");
	
	file_name = QFileDialog::getOpenFileName(this,
											 QStringLiteral("Select ballbar measurement file"),
											 d_data_path,
											 file_filters);
	
	if(file_name.length())
	{
		this->Import_Ballbar_Data(file_name);
	}
}

void TBallBarViewer::Menu_About(void)
{
	d_msg_box->setText("Help - About");
	d_msg_box->setInformativeText("Ballbar Viewer - Version 4.0");
	d_msg_box->setDetailedText(QString());
	d_msg_box->setIcon(QMessageBox::Information);
	d_msg_box->setStandardButtons(QMessageBox::Ok);
	d_msg_box->setDefaultButton(QMessageBox::Ok);
	
	d_msg_box->exec();
}

void TBallBarViewer::Tolerance_Changed(void)
{	
	this->Update_Display(false);
}

void TBallBarViewer::Ballbar_Item_Clicked(
	const int							id)
{
	double								average;
	double								median;
	double								deviation_average;
	double								deviation_median;
	QString								text;
	TBallBarPattern::TBallBarMeasurement meas;
		
	if(!(id < 0))
	{
		meas = d_ballbar_pattern.BallBar_Measurement(id);
		
		average = d_ballbar_pattern.Length_Average();
		deviation_average = meas.length - average;
		
		median = (d_ballbar_pattern.Length_Max() + d_ballbar_pattern.Length_Min())/2.0;
		deviation_median = meas.length - median;

		d_measurement_widget->Enable_Display(true);
		
		d_measurement_widget->Set_Position_Name(meas.name);
		d_measurement_widget->Set_Length(meas.length);
		d_measurement_widget->Set_Mean_Deviation(deviation_average);
		d_measurement_widget->Set_Median_Deviation(deviation_median);
		d_measurement_widget->Set_Sphere1(meas.sphere_position1,meas.sphere_form1);
		d_measurement_widget->Set_Sphere2(meas.sphere_position2,meas.sphere_form2);
		d_measurement_widget->Set_Probe(meas.probe_offset);
		
		d_opengl_display_widget->Clear_Selection();
		d_opengl_display_widget->Set_Selection(id);
	}
	else
	{
		d_measurement_widget->Enable_Display(false);
	}
}

void TBallBarViewer::dragEnterEvent(
	QDragEnterEvent 					*drag_enter_event)
{
	if(drag_enter_event->mimeData()->hasFormat("text/uri-list"))
	{
		drag_enter_event->acceptProposedAction();
	}
}

void TBallBarViewer::dropEvent(
	QDropEvent 							*drop_event)
{
	QList<QUrl>							urls;
	QString								file_name;
	
	urls = drop_event->mimeData()->urls();
	
	if(urls.size() == 1)
	{
		file_name = (*urls.begin()).toLocalFile();
		
		this->Import_Ballbar_Data(file_name);
	}
}

void TBallBarViewer::closeEvent(
	QCloseEvent							*event)
{
	QList<int>							splitter_sizes;
	
	splitter_sizes = d_splitter_widget->sizes();
	
	if(splitter_sizes.size() == 2)
	{
		d_settings->setValue("Graphic_Window_Height", splitter_sizes[0]);
		d_settings->setValue("Analysis_Window_Height", splitter_sizes[1]);
	}
	
	d_settings->setValue("Tolerance_A", d_tolerance_widget->Tolerance_A());
	d_settings->setValue("Tolerance_B", d_tolerance_widget->Tolerance_B());
	d_settings->setValue("Tolerance_C", d_tolerance_widget->Tolerance_C());
	d_settings->setValue("Tolerance_D", d_tolerance_widget->Tolerance_D());
	
	d_settings->setValue("Mainwindow_Position", this->pos());
	d_settings->setValue("Mainwindow_Size", this->size());

	event->accept();
}

void TBallBarViewer::Import_Ballbar_Data(
	const QString						&file_name)
{
	int									index;
	
	d_opengl_display_widget->Clear_BallBars();
	d_analysis_graph_widget->DeleteDataSets();
	d_analysis_text_edit->clear();
	d_tolerance_widget->Enable_Results(false);

	if(!d_ballbar_pattern.Load_Data(file_name))
	{
		d_analysis_text_edit->appendPlainText(QString("ERR:  %1").arg(d_ballbar_pattern.Last_Error()));
		return;
	}
	
	index = file_name.lastIndexOf('/');
	
	if(!(index < 0))
	{
		d_data_path = file_name;
		d_data_path.truncate(index);
	}
	
	this->Update_Display(true);
}

void TBallBarViewer::Update_Display(
	const bool							scale_to_fit)
{
	static const int					NUMBER_FREQUENCY_ITEMS = 9;
	int									set_id;
	int									count;
	int									cntr;
	TVector3							vmin,vmax;
	TVector2							pnt;
	TBallBarPattern::TBallBarMeasurement	meas;
	double								dval;
	int									flier_count;
	double								median_length;
	double								average_length;
	double								lower_limit;
	double								upper_limit;
	double								range;
	double								std_deviation;
	double								tolerance;
	double								translucent_tolerance;
	bool								oot;
	bool								translucent;
	double								measurement_window;
	int									cntr_frequency;
	int									max_cntr_frequency;
	int									frequency_array[NUMBER_FREQUENCY_ITEMS];
	static const double					ZERO_EPSILON(0.000001);

	count = d_ballbar_pattern.Measurement_Count();
	
	if(count < 1)
	{
		return;
	}
	
	average_length = d_ballbar_pattern.Length_Average();
	
	tolerance = d_tolerance_widget->Tolerance_Value(average_length);
	
	translucent_tolerance = tolerance * d_tolerance_widget->Sub_Tolerance_Percent();

	
	vmin = d_ballbar_pattern.Volume_Corner_Min();
	vmax = d_ballbar_pattern.Volume_Corner_Max();
	
	lower_limit = d_ballbar_pattern.Length_Min();
	upper_limit = d_ballbar_pattern.Length_Max();
	std_deviation = d_ballbar_pattern.Length_Stdev();

	range = upper_limit - lower_limit;
	median_length = (upper_limit + lower_limit)/2.0;

	d_opengl_display_widget->Define_Machine_Volume(vmin,vmax);
	d_tolerance_widget->Set_Results(range,median_length);

	memset(frequency_array,0,sizeof(int) * NUMBER_FREQUENCY_ITEMS);
	
	d_analysis_text_edit->clear();
	
	d_analysis_text_edit->appendPlainText(QStringLiteral("Measurement Volume"));
	d_analysis_text_edit->appendPlainText(QStringLiteral("-----------------------------------------------------------------------"));
	d_analysis_text_edit->appendPlainText(QString("          Start X: %1 mm").arg(vmin.x,10,'f',4));
	d_analysis_text_edit->appendPlainText(QString("            End X: %1 mm").arg(vmax.x,10,'f',4));
	d_analysis_text_edit->appendPlainText(QString("          Range X: %1 mm").arg(vmax.x - vmin.x,10,'f',4));
	d_analysis_text_edit->appendPlainText(QString("          Start Y: %1 mm").arg(vmin.y,10,'f',4));
	d_analysis_text_edit->appendPlainText(QString("            End Y: %1 mm").arg(vmax.y,10,'f',4));
	d_analysis_text_edit->appendPlainText(QString("          Range Y: %1 mm").arg(vmax.y - vmin.y,10,'f',4));
	d_analysis_text_edit->appendPlainText(QString("          Start Z: %1 mm").arg(vmin.z,10,'f',4));
	d_analysis_text_edit->appendPlainText(QString("            End Z: %1 mm").arg(vmax.z,10,'f',4));
	d_analysis_text_edit->appendPlainText(QString("          Range Z: %1 mm\n").arg(vmax.z - vmin.z,10,'f',4));

	d_analysis_text_edit->appendPlainText(QStringLiteral("Measurement Statistics"));
	d_analysis_text_edit->appendPlainText(QStringLiteral("-----------------------------------------------------------------------"));
	d_analysis_text_edit->appendPlainText(QString("Measurement Count: %1").arg(count,10));
	d_analysis_text_edit->appendPlainText(QString("   Minimum Length: %1").arg(lower_limit,10,'f',4));
	d_analysis_text_edit->appendPlainText(QString("   Maximum Length: %1").arg(upper_limit,10,'f',4));
	d_analysis_text_edit->appendPlainText(QString("     Range Length: %1").arg(range,10,'f',4));
	d_analysis_text_edit->appendPlainText(QString("      Mean Length: %1").arg(average_length,10,'f',4));
	d_analysis_text_edit->appendPlainText(QString("    Median Length: %1").arg(median_length,10,'f',4));
	d_analysis_text_edit->appendPlainText(QString("    StdDev Length: %1").arg(std_deviation,10,'f',4));
	d_analysis_text_edit->appendPlainText(QString("        Tolerance: %1\n").arg(tolerance,10,'f',4));

	d_analysis_text_edit->appendPlainText(QString("Estimated Range for 99% of Measurements: %1").arg(std_deviation * 3,0,'f',5));
	
	measurement_window = (upper_limit - lower_limit) / static_cast<double>(NUMBER_FREQUENCY_ITEMS - 1);
	
	d_opengl_display_widget->Clear_BallBars();

	flier_count = 0;
	for(cntr = 0;cntr < count;++cntr)
	{
		meas = d_ballbar_pattern.BallBar_Measurement(cntr);
		
		if(meas.length > (median_length + tolerance/2.0) ||
		   meas.length < (median_length - tolerance/2.0))
		{
			oot = true;
		}
		else
		{
			oot = false;
		}
		
		if(translucent_tolerance > ZERO_EPSILON)
		{
			if(meas.length > (median_length + translucent_tolerance/2.0) ||
			meas.length < (median_length - translucent_tolerance/2.0))
			{
				translucent = false;
			}
			else
			{
				translucent = true;
			}
		}
		else
		{
			translucent = false;
		}

		
		if((meas.length < (average_length - std_deviation * 1.5)) || (meas.length > (average_length + std_deviation * 1.5)))
		{
			flier_count++;
		}
		
		d_opengl_display_widget->Add_BallBar(cntr,meas.name,meas.sphere_position1,meas.sphere_position2,meas.probe_offset,oot,translucent);
	}
	
	d_opengl_display_widget->Update(scale_to_fit);
	
	d_analysis_text_edit->appendPlainText(QString("Number Outside of Estimated Range: %1\n").arg(flier_count));

	d_analysis_text_edit->appendPlainText(QStringLiteral("Individual Measurement Results"));
	d_analysis_text_edit->appendPlainText(QStringLiteral("-----------------------------------------------------------------------\n"));
	d_analysis_text_edit->appendPlainText(QStringLiteral(" Position Name          Length    Mean Dev.  Sph1 Form  Sph2 Form"));
	d_analysis_text_edit->appendPlainText(QStringLiteral("-----------------------------------------------------------------------"));

	
	for(cntr = 0;cntr < count;++cntr)
	{
		meas = d_ballbar_pattern.BallBar_Measurement(cntr);
		
		d_analysis_text_edit->appendPlainText(QString("%1 %2 %3 %4 %5")
										  .arg(meas.name,-20)
										  .arg(meas.length,10,'f',4)
										  .arg(meas.length - average_length,10,'f',4)
										  .arg(meas.sphere_form1,10,'f',4)
										  .arg(meas.sphere_form2,10,'f',4));
		
		dval = lower_limit;
		
		for(cntr_frequency = 0;cntr_frequency < NUMBER_FREQUENCY_ITEMS;cntr_frequency++)
		{
			if(meas.length > (dval - (measurement_window/2.0)) && meas.length < (dval + (measurement_window/2.0)))
			{
				frequency_array[cntr_frequency]++;
				break;
			}
			
			dval += measurement_window;
		}
	}

	if(range > 0.000001)
	{
		d_analysis_graph_widget->DeleteDataSets();
				
		set_id = d_analysis_graph_widget->CreateDataSet(QColor(0,0,255));
		
		// Force one value to zero
		pnt.x = average_length;
		pnt.y = 0;
		
		d_analysis_graph_widget->AddPoint(set_id,pnt);
		
		max_cntr_frequency = 0;
		
		for(cntr=0;cntr < NUMBER_FREQUENCY_ITEMS;++cntr)
		{
			pnt.x = lower_limit + static_cast<double>(cntr) * measurement_window;
			pnt.y = static_cast<double>(frequency_array[cntr]);
			
			if(frequency_array[cntr] > max_cntr_frequency)
			{
				max_cntr_frequency = frequency_array[cntr];
			}
			
			d_analysis_graph_widget->AddPoint(set_id,pnt);
		}
		
		// draw tolerance lines
		
		set_id = d_analysis_graph_widget->CreateDataSet(QColor(255,0,0),TGraphWidget::LINE_GRAPH);
		
		pnt.x = median_length - tolerance / 2.0;
		pnt.y = 0;
		d_analysis_graph_widget->AddPoint(set_id,pnt);
		
		pnt.y = static_cast<double>(max_cntr_frequency + 1);
		d_analysis_graph_widget->AddPoint(set_id,pnt);
		
		pnt.x = median_length + tolerance / 2.0;
		pnt.y = static_cast<double>(max_cntr_frequency + 1);
		d_analysis_graph_widget->AddPoint(set_id,pnt);
		
		pnt.y = 0;
		d_analysis_graph_widget->AddPoint(set_id,pnt);

		
		d_analysis_graph_widget->ScaleToFit();
	}
	else
	{
		d_analysis_text_edit->appendPlainText(QStringLiteral("\nINF:  There range of the data is too small to graph"));
	}

	d_analysis_text_edit->moveCursor(QTextCursor::Start);
	
	d_measurement_widget->Enable_Display(false);
}


