/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QtOpenGL/QGLWidget>
#include <QtOpenGL/QGLPixelBuffer>
#include <QOpenGLShaderProgram>
#include <QPainter>
#include <QRect>
#include <QColor>
#include <assert.h>

#include "../../core/openglshaders.h"

#include "opengldisplay.h"

static const double						ZERO_EPSILON = 0.0000001;

TOpenGLDisplay::TOpenGLDisplay(
    const QWidget						*parent,
    Qt::WindowFlags 					flags)
:TOpenGLWidget(const_cast<QWidget*>(parent),flags)
{
}

TOpenGLDisplay::~TOpenGLDisplay(void)
{
}

void TOpenGLDisplay::Define_Machine_Volume(
   const TVector3						&pnt_min,
   const TVector3						&pnt_max)
{
	TVector3							pt,vec;
		
	d_axis_draw_pos = pnt_min;
	
	d_min_xyz = pnt_min;
	d_max_xyz = pnt_max;
	
	pt = d_max_xyz - d_min_xyz;
	vec = pt;
	
	// find shortest vector
	d_axis_draw_len = pt.x;
	
	if(pt.y < d_axis_draw_len)
	{
		d_axis_draw_len = pt.y;
	}
	
	if(pt.z < d_axis_draw_len)
	{
		d_axis_draw_len = pt.z;
	}
	
	if(vec.Length() > ZERO_EPSILON)
	{
		vec.Normal();
		
		d_axis_draw_len /= 5;
		
		d_axis_draw_pos = d_min_xyz - vec * d_axis_draw_len;
		d_axis_draw_pos = d_min_xyz - vec * d_axis_draw_len;
	}
}

void TOpenGLDisplay::Clear_Selection(void)
{
	std::vector<TBallBar>::iterator		iter;
	
	for(iter = d_ballbars.begin();iter != d_ballbars.end();++iter)
	{
		(*iter).show_selected = false;
	}
	
	this->update();
}

void TOpenGLDisplay::Set_Selection(
	const int							id)
{
	std::vector<TBallBar>::iterator		iter;
	
	for(iter = d_ballbars.begin();iter != d_ballbars.end();++iter)
	{
		if((*iter).id == id)
		{
			(*iter).show_selected = true;
		}
		else
		{
			(*iter).show_selected = false;
		}

	}
	
	this->update();

}

void TOpenGLDisplay::Clear_BallBars(void)
{
	std::vector<TBallBar>::iterator		iter;
	
	for(iter = d_ballbars.begin();iter != d_ballbars.end();++iter)
	{
		d_opengl_vertex_buffer.Remove_Buffer((*iter).ballbar_shaft_vertex_id);
		d_opengl_vertex_buffer.Remove_Buffer((*iter).ballbar_shaft_normal_id);
	
		d_opengl_vertex_buffer.Remove_Buffer((*iter).ballbar_metal_vertex_id);
		d_opengl_vertex_buffer.Remove_Buffer((*iter).ballbar_metal_normal_id);

		(*iter).ballbar_shaft_vertex_id = -1;
		(*iter).ballbar_shaft_normal_id = -1;
		
		(*iter).ballbar_metal_vertex_id = -1;
		(*iter).ballbar_metal_normal_id = -1;
	}
	
	d_ballbars.clear();
	
	this->Clear_Selection_Items();
}

void TOpenGLDisplay::Add_BallBar(
	const TBallBar						&ballbar)
{
	std::vector<TOpenGLGeometryFactory::TOpenGLTriangleVertex>::const_iterator iter;
	TBallBar							new_ballbar(ballbar);
	TSelectionItemType					si;
	double								*vertex;
	double								*normal;
	unsigned int						size;
	unsigned int						ucntr;
	
	size = new_ballbar.ballbar_shaft.size() * 3;
	
	new_ballbar.ballbar_shaft_vertex_id = d_opengl_vertex_buffer.Create_Buffer(size);
	new_ballbar.ballbar_shaft_normal_id = d_opengl_vertex_buffer.Create_Buffer(size);
	
	vertex = d_opengl_vertex_buffer.Buffer(new_ballbar.ballbar_shaft_vertex_id);
	normal = d_opengl_vertex_buffer.Buffer(new_ballbar.ballbar_shaft_normal_id);
	
	if(vertex && normal)
	{
		ucntr = 0;
		for(iter = new_ballbar.ballbar_shaft.begin();iter != new_ballbar.ballbar_shaft.end();++iter)
		{
			vertex[ucntr] = (*iter).vertex.x;
			vertex[ucntr+1] = (*iter).vertex.y;
			vertex[ucntr+2] = (*iter).vertex.z;
			
			normal[ucntr] = (*iter).normal.i;
			normal[ucntr+1] = (*iter).normal.j;
			normal[ucntr+2] = (*iter).normal.k;
			
			ucntr += 3;
		}
	}
	
	size = new_ballbar.ballbar_spheres.size() * 3;
	
	new_ballbar.ballbar_metal_vertex_id = d_opengl_vertex_buffer.Create_Buffer(size);
	new_ballbar.ballbar_metal_normal_id = d_opengl_vertex_buffer.Create_Buffer(size);
	
	vertex = d_opengl_vertex_buffer.Buffer(new_ballbar.ballbar_metal_vertex_id);
	normal = d_opengl_vertex_buffer.Buffer(new_ballbar.ballbar_metal_normal_id);
	
	if(vertex && normal)
	{
		ucntr = 0;
		for(iter = new_ballbar.ballbar_spheres.begin();iter != new_ballbar.ballbar_spheres.end();++iter)
		{
			vertex[ucntr] = (*iter).vertex.x;
			vertex[ucntr+1] = (*iter).vertex.y;
			vertex[ucntr+2] = (*iter).vertex.z;
			
			normal[ucntr] = (*iter).normal.i;
			normal[ucntr+1] = (*iter).normal.j;
			normal[ucntr+2] = (*iter).normal.k;
			
			ucntr += 3;
		}
	}
	
	size = new_ballbar.probe_stem.size() * 3;
	
	new_ballbar.probe_stem_vertex_id = d_opengl_vertex_buffer.Create_Buffer(size);
	new_ballbar.probe_stem_normal_id = d_opengl_vertex_buffer.Create_Buffer(size);
	
	vertex = d_opengl_vertex_buffer.Buffer(new_ballbar.probe_stem_vertex_id);
	normal = d_opengl_vertex_buffer.Buffer(new_ballbar.probe_stem_normal_id);
	
	if(vertex && normal)
	{
		ucntr = 0;
		for(iter = new_ballbar.probe_stem.begin();iter != new_ballbar.probe_stem.end();++iter)
		{
			vertex[ucntr] = (*iter).vertex.x;
			vertex[ucntr+1] = (*iter).vertex.y;
			vertex[ucntr+2] = (*iter).vertex.z;
			
			normal[ucntr] = (*iter).normal.i;
			normal[ucntr+1] = (*iter).normal.j;
			normal[ucntr+2] = (*iter).normal.k;
			
			ucntr += 3;
		}
	}
	
	size = new_ballbar.probe_ruby.size() * 3;
	
	new_ballbar.probe_ruby_vertex_id = d_opengl_vertex_buffer.Create_Buffer(size);
	new_ballbar.probe_ruby_normal_id = d_opengl_vertex_buffer.Create_Buffer(size);
	
	vertex = d_opengl_vertex_buffer.Buffer(new_ballbar.probe_ruby_vertex_id);
	normal = d_opengl_vertex_buffer.Buffer(new_ballbar.probe_ruby_normal_id);
	
	if(vertex && normal)
	{
		ucntr = 0;
		for(iter = new_ballbar.probe_ruby.begin();iter != new_ballbar.probe_ruby.end();++iter)
		{
			vertex[ucntr] = (*iter).vertex.x;
			vertex[ucntr+1] = (*iter).vertex.y;
			vertex[ucntr+2] = (*iter).vertex.z;
			
			normal[ucntr] = (*iter).normal.i;
			normal[ucntr+1] = (*iter).normal.j;
			normal[ucntr+2] = (*iter).normal.k;
			
			ucntr += 3;
		}
	}
	
	new_ballbar.ballbar_shaft.clear();
	new_ballbar.ballbar_spheres.clear();
	new_ballbar.probe_stem.clear();
	new_ballbar.probe_ruby.clear();

	if(ballbar.show_translucent == true || d_ballbars.size() == 0)
	{
		d_ballbars.push_back(new_ballbar);
	}
	else
	{
		d_ballbars.insert(d_ballbars.begin(),new_ballbar);
	}
	
	si.selection_id = ballbar.id;
	si.min_corner = ballbar.min_xyz;
	si.max_corner = ballbar.max_xyz;
	si.selection_mode = TOpenGLWidget::SELECTION_LINE;
	
	this->AddSelectionItem(si);
}

void TOpenGLDisplay::GLInit(void)
{
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LESS);
	glEnable(GL_CULL_FACE);
	glFrontFace(GL_CCW);
	glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
	glShadeModel(GL_SMOOTH);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);

	d_triangle_shader_program = new QOpenGLShaderProgram(this);
	d_triangle_shader_program->addShaderFromSourceCode(QOpenGLShader::Vertex,OpenGLShaders::triangle_vertex_shader);
	d_triangle_shader_program->addShaderFromSourceCode(QOpenGLShader::Fragment,OpenGLShaders::triangle_fragment_shader);
	d_triangle_shader_program->link();
	
	d_triangle_uniform_color = d_triangle_shader_program->uniformLocation("color");
	d_triangle_uniform_specular = d_triangle_shader_program->uniformLocation("specular");
	d_triangle_uniform_offset = d_triangle_shader_program->uniformLocation("offset");
	d_triangle_attribute_vertex = d_triangle_shader_program->attributeLocation("vertex");
	d_triangle_attribute_normal = d_triangle_shader_program->attributeLocation("normal");
	
	d_basic_shader_program = new QOpenGLShaderProgram(this);
	d_basic_shader_program->addShaderFromSourceCode(QOpenGLShader::Vertex,OpenGLShaders::basic_vertex_shader);
	d_basic_shader_program->addShaderFromSourceCode(QOpenGLShader::Fragment,OpenGLShaders::basic_fragment_shader);
	d_basic_shader_program->link();
	
	d_basic_uniform_color = d_basic_shader_program->uniformLocation("color");
	d_basic_attribute_vertex = d_basic_shader_program->attributeLocation("vertex");

	glClearColor(1.0f,1.0f,1.0f,1.0f);
	
	// default view
	this->Rotate_Model(TVector3(1,0,0),-1.31);	// 75 degrees
	this->Rotate_Model(TVector3(0,0,1),-0.26);	// 15 degrees
}

void TOpenGLDisplay::GLResize(const int width,const int height)
{
	Q_UNUSED(width);
	Q_UNUSED(height);
}

void TOpenGLDisplay::GLRender(void)
{	
	std::vector<TBallBar>::const_iterator bbar_iter;
	TVector3							pt1,pt2;
	TVector3							axis_display_positions[3];
	TVector3							color;
	QString								text;
	QRect								rect;
	QColor								label_color(224,224,224,160);
	int									text_x,text_y;
	double								specular;
	double								opacity;
	unsigned int						ucntr;
	unsigned int						ucount;
	double								axis_lines[18];				// 3 lines, 2 points, 3 axis
	double								*vertex;

	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	glEnable(GL_DEPTH_TEST);
	glDisable(GL_LIGHTING);
	
	// X Axis display line
	
	pt2 = d_axis_draw_pos + TVector3(1,0,0) * d_axis_draw_len;
	axis_display_positions[0] = pt2 + TVector3(1,0,0) * d_axis_draw_len/8;
	axis_lines[0] = d_axis_draw_pos.x;
	axis_lines[1] = d_axis_draw_pos.y;
	axis_lines[2] = d_axis_draw_pos.z;
	axis_lines[3] = pt2.x;
	axis_lines[4] = pt2.y;
	axis_lines[5] = pt2.z;
	
	// Y Axis display line
	
	pt2 = d_axis_draw_pos + TVector3(0,1,0) * d_axis_draw_len;
	axis_display_positions[1] = pt2 + TVector3(0,1,0) * d_axis_draw_len/8;
	axis_lines[6] = d_axis_draw_pos.x;
	axis_lines[7] = d_axis_draw_pos.y;
	axis_lines[8] = d_axis_draw_pos.z;
	axis_lines[9] = pt2.x;
	axis_lines[10] = pt2.y;
	axis_lines[11] = pt2.z;
	
	// Z Axis display line
	
	pt2 = d_axis_draw_pos + TVector3(0,0,1) * d_axis_draw_len;
	axis_display_positions[2] = pt2 + TVector3(0,0,1) * d_axis_draw_len/8;
	axis_lines[12] = d_axis_draw_pos.x;
	axis_lines[13] = d_axis_draw_pos.y;
	axis_lines[14] = d_axis_draw_pos.z;
	axis_lines[15] = pt2.x;
	axis_lines[16] = pt2.y;
	axis_lines[17] = pt2.z;

	
	if(this->Is_Processing_Scale_To_Fit())
	{
		
		if(d_ballbars.size() == 0)
		{
			this->Process_Vertex(d_axis_draw_pos - TVector3(d_axis_draw_len,d_axis_draw_len,d_axis_draw_len));

			this->Process_Vertex(axis_lines[3] + d_axis_draw_len,axis_lines[4],axis_lines[5]);
			this->Process_Vertex(axis_lines[9],axis_lines[10] + d_axis_draw_len,axis_lines[11]);
			this->Process_Vertex(axis_lines[15],axis_lines[16],axis_lines[17] + d_axis_draw_len);
		}
		else
		{
			this->Process_Vertex(d_axis_draw_pos);

			this->Process_Vertex(axis_lines[3],axis_lines[4],axis_lines[5]);
			this->Process_Vertex(axis_lines[9],axis_lines[10],axis_lines[11]);
			this->Process_Vertex(axis_lines[15],axis_lines[16],axis_lines[17]);
		}
		
		for(bbar_iter = d_ballbars.begin();bbar_iter != d_ballbars.end();++bbar_iter)
		{
			if(!((*bbar_iter).ballbar_shaft_vertex_id < 0))
			{
				vertex = d_opengl_vertex_buffer.Buffer((*bbar_iter).ballbar_shaft_vertex_id);
				ucount = d_opengl_vertex_buffer.Buffer_Size((*bbar_iter).ballbar_shaft_vertex_id);
				
				if(vertex)
				{
					for(ucntr = 0;ucntr < ucount;ucntr += 3)
					{
						this->Process_Vertex(vertex[ucntr],vertex[ucntr+1],vertex[ucntr+2]);
					}
				}
			}
			
			if(!((*bbar_iter).ballbar_metal_vertex_id < 0))
			{
				vertex = d_opengl_vertex_buffer.Buffer((*bbar_iter).ballbar_metal_vertex_id);
				ucount = d_opengl_vertex_buffer.Buffer_Size((*bbar_iter).ballbar_metal_vertex_id);
				
				if(vertex)
				{
					for(ucntr = 0;ucntr < ucount;ucntr += 3)
					{
						this->Process_Vertex(vertex[ucntr],vertex[ucntr+1],vertex[ucntr+2]);
					}
				}
			}
			
			if(!((*bbar_iter).probe_stem_vertex_id < 0))
			{
				vertex = d_opengl_vertex_buffer.Buffer((*bbar_iter).probe_stem_vertex_id);
				ucount = d_opengl_vertex_buffer.Buffer_Size((*bbar_iter).probe_stem_vertex_id);
				
				if(vertex)
				{
					for(ucntr = 0;ucntr < ucount;ucntr += 3)
					{
						this->Process_Vertex(vertex[ucntr],vertex[ucntr+1],vertex[ucntr+2]);
					}
				}
			}
		}
	}
	else
	{
		d_triangle_shader_program->bind();
		
		for(bbar_iter = d_ballbars.begin();bbar_iter != d_ballbars.end();++bbar_iter)
		{
			if(!(*bbar_iter).show_translucent)
			{
				if(d_opengl_vertex_buffer.Has_Buffer((*bbar_iter).ballbar_shaft_vertex_id) && d_opengl_vertex_buffer.Has_Buffer((*bbar_iter).ballbar_shaft_normal_id))
				{
					if((*bbar_iter).show_oot == true && (*bbar_iter).show_selected == false)
					{
						color.Set(0.7,0.0,0.0);
					}
					else if((*bbar_iter).show_oot == false && (*bbar_iter).show_selected == true)
					{
						color.Set(0.0,0.0,0.7);
					}
					else if((*bbar_iter).show_oot == true && (*bbar_iter).show_selected == true)
					{
						color.Set(0.7,0.0,0.7);
					}
					else
					{
						color.Set(0.1,0.1,0.1);
					}
					
					opacity = 1.0;
					specular = 0.1;

					
					this->Render_Triangle_Buffer(d_opengl_vertex_buffer.Buffer_Size((*bbar_iter).ballbar_shaft_vertex_id),
												TVector3(0,0,0),
												color,
												specular,
												opacity,
												d_opengl_vertex_buffer.Buffer((*bbar_iter).ballbar_shaft_vertex_id),
												d_opengl_vertex_buffer.Buffer((*bbar_iter).ballbar_shaft_normal_id));

					
				}
				
				if(d_opengl_vertex_buffer.Has_Buffer((*bbar_iter).ballbar_metal_vertex_id) && d_opengl_vertex_buffer.Has_Buffer((*bbar_iter).ballbar_metal_normal_id))
				{
					if((*bbar_iter).show_oot == true && (*bbar_iter).show_selected == false)
					{
						color.Set(0.9,0.7,0.7);
					}
					else if((*bbar_iter).show_oot == false && (*bbar_iter).show_selected == true)
					{
						color.Set(0.7,0.7,0.9);
					}
					else if((*bbar_iter).show_oot == true && (*bbar_iter).show_selected == true)
					{
						color.Set(0.9,0.7,0.9);
					}
					else
					{
						color.Set(0.6,0.6,0.6);
					}
					
					opacity = 1.0;
					specular = 0.3;

					this->Render_Triangle_Buffer(d_opengl_vertex_buffer.Buffer_Size((*bbar_iter).ballbar_metal_vertex_id),
												TVector3(0,0,0),
												color,
												specular,
												opacity,
												d_opengl_vertex_buffer.Buffer((*bbar_iter).ballbar_metal_vertex_id),
												d_opengl_vertex_buffer.Buffer((*bbar_iter).ballbar_metal_normal_id));
					
				}
			}

			if((*bbar_iter).show_selected)
			{
				if(d_opengl_vertex_buffer.Has_Buffer((*bbar_iter).probe_stem_vertex_id) && d_opengl_vertex_buffer.Has_Buffer((*bbar_iter).probe_stem_normal_id))
				{
					color.Set(0.7,0.7,0.7);

					opacity = 1.0;
					specular = 0.2;

					this->Render_Triangle_Buffer(d_opengl_vertex_buffer.Buffer_Size((*bbar_iter).probe_stem_vertex_id),
												TVector3(0,0,0),
												color,
												specular,
												opacity,
												d_opengl_vertex_buffer.Buffer((*bbar_iter).probe_stem_vertex_id),
												d_opengl_vertex_buffer.Buffer((*bbar_iter).probe_stem_normal_id));
					
				}
				
				if(d_opengl_vertex_buffer.Has_Buffer((*bbar_iter).probe_ruby_vertex_id) && d_opengl_vertex_buffer.Has_Buffer((*bbar_iter).probe_ruby_normal_id))
				{
					color.Set(1.0,0.0,0.0);
					
					opacity = 1.0;
					specular = 0.25;

					this->Render_Triangle_Buffer(d_opengl_vertex_buffer.Buffer_Size((*bbar_iter).probe_ruby_vertex_id),
												TVector3(0,0,0),
												color,
												specular,
												opacity,
												d_opengl_vertex_buffer.Buffer((*bbar_iter).probe_ruby_vertex_id),
												d_opengl_vertex_buffer.Buffer((*bbar_iter).probe_ruby_normal_id));
					
				}
			}
		}
		
		glDepthFunc(GL_ALWAYS);

		for(bbar_iter = d_ballbars.begin();bbar_iter != d_ballbars.end();++bbar_iter)
		{
			if((*bbar_iter).show_translucent)
			{
				if(d_opengl_vertex_buffer.Has_Buffer((*bbar_iter).ballbar_shaft_vertex_id) && d_opengl_vertex_buffer.Has_Buffer((*bbar_iter).ballbar_shaft_normal_id))
				{
					if((*bbar_iter).show_oot == true && (*bbar_iter).show_selected == false)
					{
						color.Set(0.7,0.0,0.0);
					}
					else if((*bbar_iter).show_oot == false && (*bbar_iter).show_selected == true)
					{
						color.Set(0.0,0.0,0.7);
					}
					else if((*bbar_iter).show_oot == true && (*bbar_iter).show_selected == true)
					{
						color.Set(0.7,0.0,0.7);
					}
					else
					{
						color.Set(0.1,0.1,0.1);
					}
					
					if((*bbar_iter).show_selected)
					{
						opacity = 0.6;
					}
					else
					{
						opacity = 0.1;
					}
					specular = 0.1;
					
					color *= opacity;
					
					this->Render_Triangle_Buffer(d_opengl_vertex_buffer.Buffer_Size((*bbar_iter).ballbar_shaft_vertex_id),
												TVector3(0,0,0),
												color,
												specular,
												opacity,
												d_opengl_vertex_buffer.Buffer((*bbar_iter).ballbar_shaft_vertex_id),
												d_opengl_vertex_buffer.Buffer((*bbar_iter).ballbar_shaft_normal_id));

					
				}
				
				if(d_opengl_vertex_buffer.Has_Buffer((*bbar_iter).ballbar_metal_vertex_id) && d_opengl_vertex_buffer.Has_Buffer((*bbar_iter).ballbar_metal_normal_id))
				{
					if((*bbar_iter).show_oot == true && (*bbar_iter).show_selected == false)
					{
						color.Set(0.9,0.7,0.7);
					}
					else if((*bbar_iter).show_oot == false && (*bbar_iter).show_selected == true)
					{
						color.Set(0.7,0.7,0.9);
					}
					else if((*bbar_iter).show_oot == true && (*bbar_iter).show_selected == true)
					{
						color.Set(0.9,0.7,0.9);
					}
					else
					{
						color.Set(0.6,0.6,0.6);
					}
					
					if((*bbar_iter).show_selected)
					{
						opacity = 0.6;
					}
					else
					{
						opacity = 0.1;
					}
					specular = 0.3;
					
					color *= opacity;

					this->Render_Triangle_Buffer(d_opengl_vertex_buffer.Buffer_Size((*bbar_iter).ballbar_metal_vertex_id),
												TVector3(0,0,0),
												color,
												specular,
												opacity,
												d_opengl_vertex_buffer.Buffer((*bbar_iter).ballbar_metal_vertex_id),
												d_opengl_vertex_buffer.Buffer((*bbar_iter).ballbar_metal_normal_id));
					
				}
			}
		}
		
		glDepthFunc(GL_LESS);
		d_triangle_shader_program->release();

		d_basic_shader_program->bind();

		this->Render_Line_Buffer(18,TVector3(0.25,0.25,0.25),axis_lines);
		
		d_basic_shader_program->release();

	}
	
	
	if(!this->Is_Processing_Scale_To_Fit())
	{
		// painting of text using QPainter
		// QPainter changes one or more attributes so all relevent ones
		// are pushed then popped on end of QPainter
		glPushAttrib(GL_ACCUM_BUFFER_BIT);
		glPushAttrib(GL_VIEWPORT_BIT);
		glPushAttrib(GL_TRANSFORM_BIT);
		glPushAttrib(GL_POLYGON_BIT);
		
		glPushAttrib(GL_PIXEL_MODE_BIT);
		glPushAttrib(GL_MULTISAMPLE_BIT);
		glPushAttrib(GL_LIGHTING_BIT);
		glPushAttrib(GL_ENABLE_BIT);
		
		glPushAttrib(GL_DEPTH_BUFFER_BIT);
		glPushAttrib(GL_CURRENT_BIT);
		glPushAttrib(GL_COLOR_BUFFER_BIT);
		
		// end of openGL, start of QPainter
		QPainter							painter(this);
		
		// axis labels
		this->Point_Raster_Position(&text_x,&text_y,axis_display_positions[0]);
		text = QStringLiteral("X");
		painter.drawText(text_x,text_y,text);
		
		this->Point_Raster_Position(&text_x,&text_y,axis_display_positions[1]);
		text = QStringLiteral("Y");
		painter.drawText(text_x,text_y,text);
		
		this->Point_Raster_Position(&text_x,&text_y,axis_display_positions[2]);
		text = QStringLiteral("Z");
		painter.drawText(text_x,text_y,text);
		
		// ball bar names
		for(bbar_iter = d_ballbars.begin();bbar_iter != d_ballbars.end();++bbar_iter)
		{
			this->Point_Raster_Position(&text_x,&text_y,(*bbar_iter).name_position_space);
			
			text = (*bbar_iter).name;
			rect = painter.boundingRect(0,0,0,0,0,text);
			rect.moveTo(text_x,text_y);
			rect.adjust(-1,-1,2,2);
			
			painter.fillRect(rect,label_color);
			painter.drawText(rect,Qt::AlignHCenter | Qt::AlignVCenter,text);
		}
		
		painter.end();
		
		glPopAttrib();
		glPopAttrib();
		glPopAttrib();
		glPopAttrib();
		
		glPopAttrib();
		glPopAttrib();
		glPopAttrib();
		glPopAttrib();
		
		glPopAttrib();
		glPopAttrib();
		glPopAttrib();
	}
}


void TOpenGLDisplay::Render_Triangle_Buffer(
	const unsigned int 					size,
	const TVector3						&offset,
	const TVector3 						&color,
	const double 						&specular,
	const double						&opacity,
	double 								* const vertex,
	double 								* const nominal)
{
	d_triangle_shader_program->setUniformValue(d_triangle_uniform_color,color.x,color.y,color.z,opacity);
	d_triangle_shader_program->setUniformValue(d_triangle_uniform_offset,offset.x,offset.y,offset.z);
	d_triangle_shader_program->setUniformValue(d_triangle_uniform_specular,static_cast<float>(specular));
	
	glVertexAttribPointer(d_triangle_attribute_vertex, 3, GL_DOUBLE, GL_FALSE, 0, vertex);
	glVertexAttribPointer(d_triangle_attribute_normal, 3, GL_DOUBLE, GL_FALSE, 0, nominal);
	
	glEnableVertexAttribArray(d_triangle_attribute_vertex);
	glEnableVertexAttribArray(d_triangle_attribute_normal);
	
	glDrawArrays(GL_TRIANGLES, 0, size / 3);	// mode, first, index of 3 (x,y,z)
	
	glDisableVertexAttribArray(d_triangle_attribute_vertex);
	glDisableVertexAttribArray(d_triangle_attribute_normal);
}

void TOpenGLDisplay::Render_Line_Buffer(
	const unsigned int 					size,
	const TVector3 						&color,
	double 								* const vertex)
{
	d_basic_shader_program->setUniformValue(d_basic_uniform_color,color.x,color.y,color.z,1.0);
	
	glVertexAttribPointer(d_basic_attribute_vertex, 3, GL_DOUBLE, GL_FALSE, 0, vertex);
	
	glEnableVertexAttribArray(d_basic_attribute_vertex);
	
	glDrawArrays(GL_LINES, 0, size / 3);	// mode, first, index of 3 (x,y,z)
	
	glDisableVertexAttribArray(d_basic_attribute_vertex);
}

void TOpenGLDisplay::Render_Line_Strip_Buffer(
	const unsigned int 					size,
	const TVector3 						&color,
	double 								* const vertex)
{
	d_basic_shader_program->setUniformValue(d_basic_uniform_color,color.x,color.y,color.z,1.0);
	
	glVertexAttribPointer(d_basic_attribute_vertex, 3, GL_DOUBLE, GL_FALSE, 0, vertex);
	
	glEnableVertexAttribArray(d_basic_attribute_vertex);
	
	glDrawArrays(GL_LINE_STRIP, 0, size / 3);	// mode, first, index of 3 (x,y,z)
	
	glDisableVertexAttribArray(d_basic_attribute_vertex);
}

void TOpenGLDisplay::Render_Point_Buffer(
	const unsigned int 					size,
	const TVector3 						&color,
	double 								* const vertex)
{
	d_basic_shader_program->setUniformValue(d_basic_uniform_color,color.x,color.y,color.z,1.0);
	
	glVertexAttribPointer(d_basic_attribute_vertex, 3, GL_DOUBLE, GL_FALSE, 0, vertex);
	
	glEnableVertexAttribArray(d_basic_attribute_vertex);
	
	glDrawArrays(GL_POINTS, 0, size / 3);	// mode, first, index of 3 (x,y,z)
	
	glDisableVertexAttribArray(d_basic_attribute_vertex);
}


unsigned int TOpenGLDisplay::Convert_Geometry_To_Buffer(
	const std::vector<TOpenGLGeometryFactory::TOpenGLTriangleVertex> &data,
	double 								** const vertex,
	double 								** const normal) const
{
	std::vector<TOpenGLGeometryFactory::TOpenGLTriangleVertex>::const_iterator iter;
	int									cntr;
	unsigned int						size;
	
	if(*vertex)
	{
		delete[] (*vertex);
		(*vertex) = 0;
	}
	
	if(*normal)
	{
		delete[] (*normal);
		(*normal) = 0;
	}
	
	size = data.size() * 3;
	
	*vertex = new double[size];
	*normal = new double[size];
	
	cntr = 0;
	for(iter = data.begin();iter != data.end();++iter)
	{
		(*vertex)[cntr] = (*iter).vertex.x;
		(*vertex)[cntr+1] = (*iter).vertex.y;
		(*vertex)[cntr+2] = (*iter).vertex.z;
		
		(*normal)[cntr] = (*iter).normal.i;
		(*normal)[cntr+1] = (*iter).normal.j;
		(*normal)[cntr+2] = (*iter).normal.k;
		
		cntr += 3;
	}
	
	return size;
}





