/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QApplication>
#include <QGridLayout>
#include <QLabel>
#include <QSizePolicy>
#include <QToolButton>
#include <QFrame>
#include <QSpacerItem>
#include <QLabel>
#include <QIcon>

#include "../../core/mat4.h"
#include "../../core/openglgeometryfactory.h"

#include "ballbarpattern.h"
#include "opengldisplay.h"

#include "opengldisplaywidget.h"

static const double						ZERO_EPSILON = 0.0000001;

TOpenGLDisplayWidget::TOpenGLDisplayWidget(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QFrame(const_cast<QWidget*>(parent),flags)
{
	QGridLayout							*widget_layout;
	QSizePolicy							size_policy(QSizePolicy::Maximum,QSizePolicy::Preferred);
	QSpacerItem							*widget_vspacer;
	QString								gl_error_text;
	QString								gl_version_text;
	QString								gl_extensions_text;
	int									gl_major_version;
	
	// check version of OpenGL.  Display will not work if opengl version is less than 2
	gl_major_version = TOpenGLWidget::Get_OpenGL_Version_Major(&gl_version_text,&gl_extensions_text);
	
	size_policy.setHorizontalStretch(0);
	size_policy.setVerticalStretch(0);
	
	this->setFrameShape(QFrame::Box);
	this->setStyleSheet(QStringLiteral("background-color: rgb(255, 255, 255);"));
	
	widget_layout = new QGridLayout(this);
	widget_layout->setContentsMargins(1,1,1,1);

	if(gl_major_version > 1)
	{
		d_opengl_display = new TOpenGLDisplay(this);
		widget_layout->addWidget(d_opengl_display,0,0,6,1);
		
		d_gldisplay_error_label = 0;
	}
	else
	{
		d_gldisplay_error_label = new QLabel(this);
		d_gldisplay_error_label->setWordWrap(true);
		d_gldisplay_error_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
		widget_layout->addWidget(d_gldisplay_error_label,0,0,6,1);
		
		d_opengl_display = 0;
	}


	d_scale_to_fit_button = new QToolButton(this);
	d_scale_to_fit_button->setMaximumSize(24,24);
	d_scale_to_fit_button->setCheckable(false);
	widget_layout->addWidget(d_scale_to_fit_button,0,1,1,1);

	d_pan_button = new QToolButton(this);
	d_pan_button->setMaximumSize(24,24);
	d_pan_button->setCheckable(true);
	widget_layout->addWidget(d_pan_button,1,1,1,1);

	d_rotate_2d_button = new QToolButton(this);
	d_rotate_2d_button->setMaximumSize(24,24);
	d_rotate_2d_button->setCheckable(true);
	widget_layout->addWidget(d_rotate_2d_button,2,1,1,1);

	d_rotate_3d_button = new QToolButton(this);
	d_rotate_3d_button->setMaximumSize(24,24);
	d_rotate_3d_button->setCheckable(true);
	widget_layout->addWidget(d_rotate_3d_button,3,1,1,1);
	
	widget_vspacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
	widget_layout->addItem(widget_vspacer,4,1,1,1);

	d_scale_to_fit_button->setText(QStringLiteral("..."));
	d_pan_button->setText(QStringLiteral("..."));
	d_rotate_2d_button->setText(QStringLiteral("..."));
	d_rotate_3d_button->setText(QStringLiteral("..."));
	
	d_scale_to_fit_button->setIcon(QIcon(":/toolbar/tb_scale_to_fit.png"));
	d_pan_button->setIcon(QIcon(":/toolbar/tb_pan.png"));
	d_rotate_2d_button->setIcon(QIcon(":/toolbar/tb_rotate2d.png"));
	d_rotate_3d_button->setIcon(QIcon(":/toolbar/tb_rotate3d.png"));
	
	d_scale_to_fit_button->setToolTip(QStringLiteral("Scale to Fit"));
	d_pan_button->setToolTip(QStringLiteral("Pan Mode"));
	d_rotate_2d_button->setToolTip(QStringLiteral("Rotate 2D Mode"));
	d_rotate_3d_button->setToolTip(QStringLiteral("Rotate 3D Mode"));
	
	
	if(d_gldisplay_error_label)
	{
		gl_error_text = QStringLiteral("<html><head/><body>");
		gl_error_text += QStringLiteral("<p><span style=\" font-weight:700;\">ERR:  OpenGL version not supported.  No graphical display will be shown.</span></p>");
		gl_error_text += QStringLiteral("<p><span style=\" font-weight:500; text-decoration: underline;\">OpenGL Version Text:</span></p>");
		gl_error_text += QString("<p>%1</p>").arg(gl_version_text);
		gl_error_text += QStringLiteral("<p><span style=\" font-weight:500; text-decoration: underline;\">OpenGL Extensions:</span></p>");
		gl_error_text += QString("<p>%1</p>").arg(gl_extensions_text);
		gl_error_text += QStringLiteral("</body></html>");
		
		d_gldisplay_error_label->setText(gl_error_text);
	}
	
	this->Define_Machine_Volume(TVector3(-100,-100,-100),TVector3(100,100,100));
	
	d_pan_button->setChecked(true);
	
	connect(d_scale_to_fit_button,SIGNAL(clicked(void)),this,SLOT(Scale_To_Fit(void)));
	connect(d_pan_button,SIGNAL(toggled(bool)),this,SLOT(Pan(bool)));
	connect(d_rotate_2d_button,SIGNAL(toggled(bool)),this,SLOT(Rotate2d(bool)));
	connect(d_rotate_3d_button,SIGNAL(toggled(bool)),this,SLOT(Rotate3d(bool)));
	
	if(d_opengl_display)
	{
		connect(d_opengl_display,&TOpenGLWidget::Model_Click_Point,this,&TOpenGLDisplayWidget::Model_Item_Clicked);
	}
}

TOpenGLDisplayWidget::~TOpenGLDisplayWidget(void)
{
}

void TOpenGLDisplayWidget::Define_Machine_Volume(
	const TVector3						&pnt_min,
	const TVector3						&pnt_max)
{
	double								dval;
	
	if(!d_opengl_display)
	{
		return;
	}
	
	d_max_xyz = pnt_max;
	d_min_xyz = pnt_min;
	
	if(d_min_xyz.x > d_max_xyz.x)
	{
		dval = d_max_xyz.x;
		d_max_xyz.x = d_min_xyz.x;
		d_min_xyz.x = dval;
	}
	
	if(d_min_xyz.y > d_max_xyz.y)
	{
		dval = d_max_xyz.y;
		d_max_xyz.y = d_min_xyz.y;
		d_min_xyz.y = dval;
	}
	
	if(d_min_xyz.z > d_max_xyz.z)
	{
		dval = d_max_xyz.z;
		d_max_xyz.z = d_min_xyz.z;
		d_min_xyz.z = dval;
	}

	d_opengl_display->Define_Machine_Volume(d_min_xyz,d_max_xyz);

}

void TOpenGLDisplayWidget::Add_BallBar(
	const int							id,
	const QString						&name,
	const TVector3						&sph1,
	const TVector3						&sph2,
	const TVector3						&probe,
	const bool							oot,
	const bool							translucent)
{
	std::vector<TOpenGLGeometryFactory::TOpenGLTriangleVertex> triangles;
	TOpenGLDisplay::TBallBar			ballbar;
	TVector3							axis;
	TVector3							center;
	TVector3							p1,p2,p3,pnt;
	double								length,length_xy,length_z;
	TOpenGLGeometryFactory				geometry;
	const double						PROBE_OFFSET_DISTANCE(30);
	const double						PROBE_STYLUS_RADIUS(2.5);
	const double						MIN_PROBE_OFFSET_FOR_KNUCKLE(10);
	
	if(!d_opengl_display)
	{
		return;
	}
	
	ballbar.show_oot = oot;
	ballbar.show_translucent = translucent;
	ballbar.show_selected = false;

	geometry.Set_Geometry_Resolution(3);
	
	axis = sph1 - sph2;
	
	length = axis.Length();
	
	if(!(length > ZERO_EPSILON))
	{
		return;
	}
	
	axis.Normal();
	center = (sph1 + sph2) / 2.0;
	
	if(length > 148)	// minimum length or it won't look right
	{
		ballbar.ballbar_spheres = geometry.RenderSphere(sph1,19.5);
		
		pnt = sph1 - axis * 34.75;
		triangles = geometry.RenderCylinder(pnt,axis,10,50);
		ballbar.ballbar_spheres.insert(ballbar.ballbar_spheres.end(),triangles.begin(),triangles.end());

		pnt = sph1 - axis * 66.75;
		triangles = geometry.RenderPolygon(pnt,axis,TVector3(0,1,0),15,14,6);
		ballbar.ballbar_spheres.insert(ballbar.ballbar_spheres.end(),triangles.begin(),triangles.end());

		ballbar.ballbar_shaft = geometry.RenderCylinder(center,axis,14,length - 147.52);
		
		pnt = sph2 + axis * 66.75;
		triangles = geometry.RenderPolygon(pnt,axis,TVector3(0,1,0),15,14,6);
		ballbar.ballbar_spheres.insert(ballbar.ballbar_spheres.end(),triangles.begin(),triangles.end());

		pnt = sph2 + axis * 34.75;
		triangles = geometry.RenderCylinder(pnt,axis,10,50);
		ballbar.ballbar_spheres.insert(ballbar.ballbar_spheres.end(),triangles.begin(),triangles.end());

		triangles = geometry.RenderSphere(sph2,19.5);
		ballbar.ballbar_spheres.insert(ballbar.ballbar_spheres.end(),triangles.begin(),triangles.end());
	}
	else
	{
		// draw a shorter version
		ballbar.ballbar_spheres = geometry.RenderSphere(sph1,19.5);
		
		ballbar.ballbar_shaft = geometry.RenderCylinder(center,axis,10,length-15);
		
		triangles = geometry.RenderSphere(sph2,19.5);
		ballbar.ballbar_spheres.insert(ballbar.ballbar_spheres.end(),triangles.begin(),triangles.end());
	}
	
	// draw stylus
	p1 = p2 = probe;
	p1.z = 0;
	p2.x = 0;
	p2.y = 0;
	
	length_xy = p1.Length();
	length_z = p2.Length();
	
	if(length_xy > MIN_PROBE_OFFSET_FOR_KNUCKLE)
	{
		p1.Normal();
		
		p3 = sph1 - p1 * PROBE_OFFSET_DISTANCE;
		ballbar.probe_ruby = geometry.RenderSphere(p3,PROBE_STYLUS_RADIUS);
		
		pnt = p3 - p1 * length_xy/2.0;
		ballbar.probe_stem = geometry.RenderCylinder(pnt,p1,1.5,length_xy);

		pnt = p3 - p1 * length_xy;
		triangles = geometry.RenderPolygon(pnt,TVector3(0,0,1),p1,5,5,8);
		ballbar.probe_stem.insert(ballbar.probe_stem.end(),triangles.begin(),triangles.end());
	
		pnt.z += length_z/2.0;
		triangles = geometry.RenderCylinder(pnt,TVector3(0,0,1),1.5,length_z);
		ballbar.probe_stem.insert(ballbar.probe_stem.end(),triangles.begin(),triangles.end());
		
	}
	else if(length_z > ZERO_EPSILON)
	{
		p2.Normal();
		
		p3 = sph1 - p2 * PROBE_OFFSET_DISTANCE;
		ballbar.probe_ruby = geometry.RenderSphere(p3,PROBE_STYLUS_RADIUS);
		
		pnt = p3 - p2 * length_z/2.0;
		ballbar.probe_stem = geometry.RenderCylinder(pnt,p2,1.5,length_z);
	}
	
	
	ballbar.id = id;
	
	p1 = axis;
	p2 = p1 * TVector3(0,0,1);
	
	if(p2.Length() < ZERO_EPSILON)
	{
		p2 = p1 * TVector3(0,1,0);
	}
	
	p2.Normal();
	
	p3 = p1 * p2;
	p2 = p3 * p1;
	
	ballbar.min_xyz = sph1 + axis * 21;	// increase by just over half of sphere radius
	ballbar.max_xyz = sph2 - axis * 21;		// increase by just over half of sphere radius
	
	ballbar.name = name;
	
	// calculate where name and leader line is drawn
	
	ballbar.name_position_bar = (center + sph1)/2.0;
	
	// find right angle vector in XY plane
	p1 = axis * TVector3(0,0,1);
	
	if(p1.Length() < ZERO_EPSILON)
	{
		p1 = axis * TVector3(0,1,0);
	}
	
	p1.Normal();
	
	// second label is 10% of ball bar length
	length_xy = length * 0.10;
	
	if(length_xy < 20)
	{
		length_xy = 20;	// minimum 20 mm
	}
	
	ballbar.name_position_space = ballbar.name_position_bar + p1 * length_xy;
	
	d_opengl_display->Add_BallBar(ballbar);
}

void TOpenGLDisplayWidget::Clear_Selection(void)
{
	if(d_opengl_display)
	{
		d_opengl_display->Clear_Selection();
	}
}

void TOpenGLDisplayWidget::Set_Selection(
	const int							id)
{
	if(d_opengl_display)
	{
		d_opengl_display->Set_Selection(id);
	}
}

void TOpenGLDisplayWidget::Clear_BallBars(void)
{
	if(d_opengl_display)
	{
		d_opengl_display->Clear_BallBars();
	}
}

void TOpenGLDisplayWidget::Update(
	const bool							scale_to_fit)
{
	if(!d_opengl_display)
	{
		return;
	}
	
	d_opengl_display->update();

	if(scale_to_fit)
	{
		d_opengl_display->Scale_To_Fit();
	}
}

void TOpenGLDisplayWidget::Model_Item_Clicked(
	const double						&x,
	const double						&y,
	const double						&z,
	const int							selection_id)
{
	Q_UNUSED(x);
	Q_UNUSED(y);
	Q_UNUSED(z);
		
	emit Ballbar_Item_Clicked(selection_id);
}

void TOpenGLDisplayWidget::Scale_To_Fit(void)
{
	if(d_opengl_display)
	{
		d_opengl_display->Scale_To_Fit();
	}
}

void TOpenGLDisplayWidget::Rotate2d(
	bool								toggled)
{
	bool								prev_state;
	
	if(toggled)
	{
		prev_state = d_pan_button->blockSignals(true);
		d_pan_button->setChecked(false);
		d_pan_button->blockSignals(prev_state);
	
		prev_state = d_rotate_3d_button->blockSignals(true);
		d_rotate_3d_button->setChecked(false);
		d_rotate_3d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Rotate_2D();
		}
	}
	else
	{
		prev_state = d_pan_button->blockSignals(true);
		d_pan_button->setChecked(true);
		d_pan_button->blockSignals(prev_state);
		
		prev_state = d_rotate_2d_button->blockSignals(true);
		d_rotate_2d_button->setChecked(false);
		d_rotate_2d_button->blockSignals(prev_state);
		
		prev_state = d_rotate_3d_button->blockSignals(true);
		d_rotate_3d_button->setChecked(false);
		d_rotate_3d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Pan();
		}
	}
}

void TOpenGLDisplayWidget::Rotate3d(
	bool								toggled)
{
	bool								prev_state;
	
	if(toggled)
	{
		prev_state = d_pan_button->blockSignals(true);
		d_pan_button->setChecked(false);
		d_pan_button->blockSignals(prev_state);
		
		prev_state = d_rotate_2d_button->blockSignals(true);
		d_rotate_2d_button->setChecked(false);
		d_rotate_2d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Rotate_3D();
		}
	}
	else
	{
		prev_state = d_pan_button->blockSignals(true);
		d_pan_button->setChecked(true);
		d_pan_button->blockSignals(prev_state);
		
		prev_state = d_rotate_2d_button->blockSignals(true);
		d_rotate_2d_button->setChecked(false);
		d_rotate_2d_button->blockSignals(prev_state);
		
		prev_state = d_rotate_3d_button->blockSignals(true);
		d_rotate_3d_button->setChecked(false);
		d_rotate_3d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Pan();
		}
	}
}

void TOpenGLDisplayWidget::Pan(
	bool								toggled)
{
	bool								prev_state;
	
	if(toggled)
	{
		prev_state = d_rotate_3d_button->blockSignals(true);
		d_rotate_3d_button->setChecked(false);
		d_rotate_3d_button->blockSignals(prev_state);
		
		prev_state = d_rotate_2d_button->blockSignals(true);
		d_rotate_2d_button->setChecked(false);
		d_rotate_2d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Pan();
		}
	}
	else
	{
		prev_state = d_pan_button->blockSignals(true);
		d_pan_button->setChecked(true);
		d_pan_button->blockSignals(prev_state);
		
		prev_state = d_rotate_2d_button->blockSignals(true);
		d_rotate_2d_button->setChecked(false);
		d_rotate_2d_button->blockSignals(prev_state);
		
		prev_state = d_rotate_3d_button->blockSignals(true);
		d_rotate_3d_button->setChecked(false);
		d_rotate_3d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Pan();
		}
	}
}

