/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2024  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////


#include <QApplication>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QListWidget>
#include <QListWidgetItem>
#include <QToolButton>
#include <QSlider>
#include <QFrame>
#include <QVariant>
#include <QRegularExpression>
#include <QRegularExpressionMatch>
#include <QSpacerItem>
#include <QIcon>
#include <QPixmap>

#include "tolerancehelpdialog.h"

#include "tolerancewidget.h"

TToleranceWidget::TToleranceWidget(
	const QWidget						*parent,
	Qt::WindowFlags 					flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QGridLayout							*widget_layout;
	QGridLayout							*results_widget_layout;
	QHBoxLayout							*tolerance_hlayout;
	QLabel								*measured_range_label;
	QLabel								*sub_tolerance_label;
	QLabel 								*expression_label;
	QLabel								*calculated_tolerance_label;
	QLabel								*calculated_tolerance_units_label;
	QLabel								*calculated_sub_tolerance_label;
	QLabel								*calculated_sub_tolerance_units_label;
	QLabel								*measured_range_units_label;
	QSpacerItem							*widget_vspacer;
	QSpacerItem							*button_hspacer;
	
	widget_layout = new QGridLayout(this);
	widget_layout->setSpacing(4);
	widget_layout->setContentsMargins(4,4,4,4);

	expression_label = new QLabel(this);
	widget_layout->addWidget(expression_label,0,0,1,3);
	
	d_tolerance_edit = new QLineEdit(this);
	d_tolerance_edit->setAlignment(Qt::AlignCenter);
	widget_layout->addWidget(d_tolerance_edit,1,0,1,3);

	tolerance_hlayout = new QHBoxLayout();
	
	button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	tolerance_hlayout->addItem(button_hspacer);

	d_apply_button = new QToolButton(this);
	d_apply_button->setSizePolicy(QSizePolicy::Preferred,QSizePolicy::Preferred);
	tolerance_hlayout->addWidget(d_apply_button);

	d_help_button = new QToolButton(this);
	d_help_button->setSizePolicy(QSizePolicy::Preferred,QSizePolicy::Preferred);
	tolerance_hlayout->addWidget(d_help_button);
	widget_layout->addLayout(tolerance_hlayout,2,0,1,3);
	
	d_results_widget = new QWidget(this);
	
	results_widget_layout = new QGridLayout(d_results_widget);
	results_widget_layout->setSpacing(4);
	results_widget_layout->setContentsMargins(0,0,0,0);
	
	measured_range_label = new QLabel(d_results_widget);
	results_widget_layout->addWidget(measured_range_label,0,0,1,1);

	d_measured_range_edit = new QLineEdit(d_results_widget);
	d_measured_range_edit->setAlignment(Qt::AlignCenter);
	d_measured_range_edit->setReadOnly(true);
	results_widget_layout->addWidget(d_measured_range_edit,0,1,1,1);
	
	measured_range_units_label = new QLabel(d_results_widget);
	results_widget_layout->addWidget(measured_range_units_label,0,2,1,1);

	calculated_tolerance_label = new QLabel(d_results_widget);
	results_widget_layout->addWidget(calculated_tolerance_label,1,0,1,1);

	d_calculated_tolerance_edit = new QLineEdit(d_results_widget);
	d_calculated_tolerance_edit->setAlignment(Qt::AlignCenter);
	d_calculated_tolerance_edit->setReadOnly(true);
	results_widget_layout->addWidget(d_calculated_tolerance_edit,1,1,1,1);
	
	calculated_tolerance_units_label = new QLabel(d_results_widget);
	results_widget_layout->addWidget(calculated_tolerance_units_label,1,2,1,1);

	sub_tolerance_label = new QLabel(d_results_widget);
	results_widget_layout->addWidget(sub_tolerance_label,2,0,1,1);

	d_sub_tolerance_slider = new QSlider(d_results_widget);
	d_sub_tolerance_slider->setOrientation(Qt::Horizontal);
	d_sub_tolerance_slider->setRange(0,20);
	d_sub_tolerance_slider->setValue(0);
	results_widget_layout->addWidget(d_sub_tolerance_slider,2,1,1,1);
	
	calculated_sub_tolerance_label = new QLabel(d_results_widget);
	results_widget_layout->addWidget(calculated_sub_tolerance_label,3,0,1,1);

	d_calculated_sub_tolerance_edit = new QLineEdit(d_results_widget);
	d_calculated_sub_tolerance_edit->setAlignment(Qt::AlignCenter);
	d_calculated_sub_tolerance_edit->setReadOnly(true);
	results_widget_layout->addWidget(d_calculated_sub_tolerance_edit,3,1,1,1);
	
	calculated_sub_tolerance_units_label = new QLabel(d_results_widget);
	results_widget_layout->addWidget(calculated_sub_tolerance_units_label,3,2,1,1);

	widget_layout->addWidget(d_results_widget,3,0,1,3);
	
	widget_vspacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
	widget_layout->addItem(widget_vspacer,4,0,1,3);
	
	d_tolerance_help_dialog = new TToleranceHelpDialog(this);

	// defaults
	d_tolerance_edit->setText("0.012 + 0.006L/1000");
	d_tolerance_type = TToleranceWidget::TOLERANCE_FORMULA;
	d_apply_button->setEnabled(false);
	d_results_widget->setEnabled(false);
	d_nominal_length = 0.0;

	expression_label->setText(QStringLiteral("Tolerance Expression:"));
	calculated_tolerance_label->setText(QStringLiteral("Tolerance:"));
	calculated_tolerance_units_label->setText(QStringLiteral("mm"));
	measured_range_label->setText(QStringLiteral("Measured Range:"));
	measured_range_units_label->setText(QStringLiteral("mm"));
	sub_tolerance_label->setText(QStringLiteral("Sub Tolerance:"));
	calculated_sub_tolerance_label->setText(QStringLiteral("Sub Tolerance Range:"));
	calculated_sub_tolerance_units_label->setText(QStringLiteral("mm"));
	d_apply_button->setText(QStringLiteral("Apply"));
	d_help_button->setText(QStringLiteral("Help"));
	
	connect(d_tolerance_edit,&QLineEdit::textChanged,this,&TToleranceWidget::Update_Tolerance);
	connect(d_apply_button,&QToolButton::clicked,this,&TToleranceWidget::Apply_Tolerance);
	connect(d_help_button,&QToolButton::clicked,this,&TToleranceWidget::Show_Tolerance_Help);
	connect(d_sub_tolerance_slider,&QSlider::valueChanged,this,&TToleranceWidget::Change_Sub_Tolerance);
}

TToleranceWidget::~TToleranceWidget(void)
{
}

double TToleranceWidget::Tolerance_Value(
	const double						&length) const
{
	double								tolerance(0.0);
	
	switch(d_tolerance_type)
	{
		case TToleranceWidget::TOLERANCE_ERROR:
			break;
			
		case TToleranceWidget::TOLERANCE_BAND:
			tolerance = d_tolerance_a;
			break;
			
		case TToleranceWidget::TOLERANCE_FORMULA:
			tolerance = d_tolerance_a;
			
			if(d_tolerance_c > 0.1)
			{
				tolerance += d_tolerance_b * length / d_tolerance_c;
			}
			break;
			
		case TToleranceWidget::TOLERANCE_FORMULA_WITH_LIMIT:
			tolerance = d_tolerance_a;
			
			if(d_tolerance_c > 0.1)
			{
				tolerance += d_tolerance_b * length / d_tolerance_c;
			}
			
			if(tolerance > d_tolerance_d)
			{
				tolerance = d_tolerance_d;
			}
			break;
	}
	
	return tolerance;
}

double TToleranceWidget::Sub_Tolerance_Percent(void) const
{
	double								percent(0.0);
	double								range;
	static const double					MIN_RANGE(0.25);
	
	range = static_cast<double>(d_sub_tolerance_slider->maximum() - d_sub_tolerance_slider->minimum());
	
	if(!(range < MIN_RANGE))
	{
		percent = static_cast<double>(d_sub_tolerance_slider->value()) / range;
	}
	
	return percent;
}

double TToleranceWidget::Tolerance_Limit_Position(
	const double 						&tol_a,
	const double 						&tol_b,
	const double 						&tol_c,
	const double 						&tol_d) const
{
	double								slope;
	double								transition_position(0);
	
	if(tol_d < tol_a)
	{
		return transition_position;
	}
	
	slope = tol_b / tol_c;	// mm / m
	
	transition_position = (tol_d - tol_a) / slope;
	
	return transition_position;
}

void TToleranceWidget::Update(void)
{
	QString								tolerance_text;
	static const double					MINIMUM_ENTRY(0.00001);
	static const double					MINIMUM_LENGTH(1.0);
	
	d_tolerance_type = TToleranceWidget::TOLERANCE_ERROR;
	d_apply_button->setEnabled(false);
	
	tolerance_text = QString("%1").arg(d_tolerance_a,0,'f',4);
	
	if(d_tolerance_b > MINIMUM_ENTRY && d_tolerance_c > MINIMUM_LENGTH)
	{
		tolerance_text += QString("+%1L/%2").arg(d_tolerance_b,0,'f',4).arg(d_tolerance_c,0,'f',1);
	}
	
	if(d_tolerance_d > MINIMUM_ENTRY)
	{
		tolerance_text += QString("|%1").arg(d_tolerance_d,0,'f',4);
	}
	
	if(this->Analyze_Tolerance(tolerance_text))
	{
		d_tolerance_edit->setStyleSheet(QString());
	}
	else
	{
		d_tolerance_edit->setStyleSheet(QStringLiteral("background-color: rgb(255, 151, 160);"));
	}
	
	this->Update_Display();
}

void TToleranceWidget::Enable_Results(
	const bool 							state)
{
	d_results_widget->setEnabled(state);

	if(!state)
	{
		d_measured_range_edit->setText(QString());
		d_calculated_tolerance_edit->setText(QString());
		d_calculated_sub_tolerance_edit->setText(QString());
		d_nominal_length = 0.0;
	}
}

void TToleranceWidget::Set_Results(
	const double 						&range,
	const double						&nominal_length)
{	
	double								tolerance;
	double								dval;
	
	d_results_widget->setEnabled(true);	
	d_nominal_length = nominal_length;
	
	d_measured_range_edit->setText(QString("%1").arg(range,0,'f',4));
	
	tolerance = this->Tolerance_Value(nominal_length);
	d_calculated_tolerance_edit->setText(QString("%1").arg(tolerance,0,'f',4));
	
	dval = this->Sub_Tolerance_Percent();
	tolerance *= dval;
	d_calculated_sub_tolerance_edit->setText(QString("%1").arg(tolerance,0,'f',4));
}

void TToleranceWidget::Update_Tolerance(void)
{
	d_apply_button->setEnabled(this->Analyze_Tolerance(d_tolerance_edit->text()));
	
	d_tolerance_edit->setStyleSheet(QString());
	if(d_tolerance_type == TToleranceWidget::TOLERANCE_ERROR)
	{
		d_tolerance_edit->setStyleSheet(QStringLiteral("background-color: rgb(255, 151, 160);"));
	}
}

void TToleranceWidget::Apply_Tolerance(void)
{
	emit Tolerance_Changed();
	d_apply_button->setEnabled(false);
}

void TToleranceWidget::Change_Sub_Tolerance(void)
{
	if(!d_apply_button->isEnabled())
	{
		emit Sub_Tolerance_Changed();
	}
}

void TToleranceWidget::Show_Tolerance_Help(void)
{	
	d_tolerance_help_dialog->exec();
}

bool TToleranceWidget::Analyze_Tolerance(
	const QString 						&tolerance_text)
{
	QRegularExpression					single_regexp("^([0-9.]{1,8})$");
	QRegularExpression					expression_regexp("^([0-9.]{1,8})[ ]{0,1}[+][ ]{0,1}([0-9.]{1,8})[ ]{0,1}[L,l]{1}[ ]{0,1}[/][ ]{0,1}([0-9.]{1,9})$");
	QRegularExpression					expression_limit_regexp("^([0-9.]{1,8})[ ]{0,1}[+][ ]{0,1}([0-9.]{1,8})[ ]{0,1}[L,l]{1}[ ]{0,1}[/][ ]{0,1}([0-9.]{1,9})[ ]{0,1}[|][ ]{0,1}([0-9.]{1,8})$");
	QRegularExpressionMatch				expression_match;
	QStringList							expression_text_list;
	QString								text(tolerance_text.simplified());
	
	expression_match = single_regexp.match(text);
	if(expression_match.hasMatch())
	{
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 2)
		{
			d_tolerance_type = TToleranceWidget::TOLERANCE_BAND;
			
			d_tolerance_a = expression_text_list[1].toDouble();
			d_tolerance_b = 0;
			d_tolerance_c = 0;
			d_tolerance_d = 0;
			
			if(d_tolerance_a > 1)
			{
				d_tolerance_type = TToleranceWidget::TOLERANCE_ERROR;
				return false;
			}
			
			return true;
		}
	}
	
	expression_match = expression_regexp.match(text);
	if(expression_match.hasMatch())
	{
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 4)
		{
			d_tolerance_type = TToleranceWidget::TOLERANCE_FORMULA;
			
			d_tolerance_a = expression_text_list[1].toDouble();
			d_tolerance_b = expression_text_list[2].toDouble();
			d_tolerance_c = expression_text_list[3].toDouble();
			d_tolerance_d = 0;
			
			if(d_tolerance_a > 1 || d_tolerance_b > 1)
			{
				d_tolerance_type = TToleranceWidget::TOLERANCE_ERROR;
				return false;
			}
			
			return true;
		}
	}

	
	expression_match = expression_limit_regexp.match(text);
	if(expression_match.hasMatch())
	{
		expression_text_list = expression_match.capturedTexts();
		
		if(expression_text_list.size() == 5)
		{
			d_tolerance_type = TToleranceWidget::TOLERANCE_FORMULA_WITH_LIMIT;
			
			d_tolerance_a = expression_text_list[1].toDouble();
			d_tolerance_b = expression_text_list[2].toDouble();
			d_tolerance_c = expression_text_list[3].toDouble();
			d_tolerance_d = expression_text_list[4].toDouble();
			
			if(d_tolerance_a > 1 || d_tolerance_b > 1 || d_tolerance_d > 1)
			{
				d_tolerance_type = TToleranceWidget::TOLERANCE_ERROR;
				return false;
			}
			
			return true;
		}
	}

	d_tolerance_type = TToleranceWidget::TOLERANCE_ERROR;
	return false;
}

void TToleranceWidget::Update_Display(void)
{
	bool								prev_state;
	
	prev_state = d_tolerance_edit->blockSignals(true);
	
	switch(d_tolerance_type)
	{
		case TToleranceWidget::TOLERANCE_ERROR:
			break;
			
		case TToleranceWidget::TOLERANCE_BAND:
			d_tolerance_edit->setText(QString("%1").arg(d_tolerance_a,0,'f',4));
			break;
			
		case TToleranceWidget::TOLERANCE_FORMULA:
			d_tolerance_edit->setText(QString("%1+%2L/%3")
									  .arg(d_tolerance_a,0,'f',4)
									  .arg(d_tolerance_b,0,'f',4)
									  .arg(d_tolerance_c,0,'f',0));
			break;
			
		case TToleranceWidget::TOLERANCE_FORMULA_WITH_LIMIT:
			d_tolerance_edit->setText(QString("%1+%2L/%3 | %4")
									  .arg(d_tolerance_a,0,'f',4)
									  .arg(d_tolerance_b,0,'f',4)
									  .arg(d_tolerance_c,0,'f',0)
									  .arg(d_tolerance_d,0,'f',4));
			break;
	}
	
	d_tolerance_edit->blockSignals(prev_state);
}


